from typing import Optional

import httpx
from fastapi import Depends, HTTPException, status
from fastapi_users.authentication import BaseAuthentication, CookieAuthentication  # type: ignore
from fastapi_users import FastAPIUsers, BaseUserManager  # type: ignore
from starlette.requests import Request
from fps.exceptions import RedirectException  # type: ignore

from .models import (
    User,
    UserCreate,
    UserUpdate,
    UserDB,
)
from .config import get_auth_config
from .db import secret, get_user_db


NOAUTH_EMAIL = "noauth_user@jupyter.com"
NOAUTH_USER = UserDB(
    id="d4ded46b-a4df-4b51-8d83-ae19010272a7",
    email=NOAUTH_EMAIL,
    hashed_password="",
)


class NoAuthAuthentication(BaseAuthentication):
    def __init__(self, user: UserDB, name: str = "noauth"):
        super().__init__(name, logout=False)
        self.user = user
        self.scheme = None  # type: ignore

    async def __call__(self, credentials, user_manager):
        noauth_user = await user_manager.user_db.get_by_email(NOAUTH_EMAIL)
        return noauth_user or self.user


noauth_authentication = NoAuthAuthentication(NOAUTH_USER)


class UserManager(BaseUserManager[UserCreate, UserDB]):
    user_db_model = UserDB

    async def on_after_register(self, user: UserDB, request: Optional[Request] = None):
        user.initialized = True
        for oauth_account in user.oauth_accounts:
            if oauth_account.oauth_name == "github":
                async with httpx.AsyncClient() as client:
                    r = (
                        await client.get(
                            f"https://api.github.com/user/{oauth_account.account_id}"
                        )
                    ).json()
                user.anonymous = False
                user.username = r["login"]
                user.name = r["name"]
                user.color = None
                user.avatar = r["avatar_url"]
        await self.user_db.update(user)


class LoginCookieAuthentication(CookieAuthentication):
    async def get_login_response(self, user, response, user_manager):
        await super().get_login_response(user, response, user_manager)
        # set user as logged in
        user.logged_in = True
        await user_manager.user_db.update(user)
        # auto redirect
        response.status_code = status.HTTP_302_FOUND
        response.headers["Location"] = "/lab"

    async def get_logout_response(self, user, response, user_manager):
        await super().get_logout_response(user, response, user_manager)
        # set user as logged out
        user.logged_in = False
        await user_manager.user_db.update(user)


cookie_authentication = LoginCookieAuthentication(
    cookie_secure=get_auth_config().cookie_secure, secret=secret
)


def get_user_manager(user_db=Depends(get_user_db)):
    yield UserManager(user_db)


users = FastAPIUsers(
    get_user_manager,
    [noauth_authentication, cookie_authentication],
    User,
    UserCreate,
    UserUpdate,
    UserDB,
)


async def get_enabled_backends(auth_config=Depends(get_auth_config)):
    if auth_config.mode == "noauth":
        return [noauth_authentication]
    return [cookie_authentication]


def current_user():
    async def _(
        user: User = Depends(
            users.current_user(optional=True, get_enabled_backends=get_enabled_backends)
        ),
        auth_config=Depends(get_auth_config),
    ):
        if user is None:
            if auth_config.login_url:
                raise RedirectException(auth_config.login_url)
            else:
                raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED)
        else:
            return user

    return _



from uuid import uuid4

from httpx_oauth.clients.github import GitHubOAuth2  # type: ignore
from fps.hooks import register_router  # type: ignore
from fps.config import get_config  # type: ignore
from fps.logging import get_configured_logger  # type: ignore
from fps_uvicorn.config import UvicornConfig  # type: ignore
from fastapi import APIRouter, Depends
from sqlalchemy.orm import sessionmaker  # type: ignore

from .config import get_auth_config
from .db import user_db, secret, database, engine, UserTable
from .backends import (
    users,
    current_user,
    cookie_authentication,
    NOAUTH_USER,
    NOAUTH_EMAIL,
)
from .models import (
    User,
    UserDB,
)

logger = get_configured_logger("auth")

Session = sessionmaker(bind=engine)
session = Session()

uvicorn_config = get_config(UvicornConfig)
auth_config = get_auth_config()


github_oauth_client = GitHubOAuth2(
    auth_config.client_id, auth_config.client_secret.get_secret_value()
)


github_oauth_router = users.get_oauth_router(
    github_oauth_client, secret  # type: ignore
)
auth_router = users.get_auth_router(cookie_authentication)
user_register_router = users.get_register_router()  # type: ignore
users_router = users.get_users_router()

router = APIRouter()

TOKEN_USER = None
USER_TOKEN = None


def set_user_token(user_token):
    global USER_TOKEN
    USER_TOKEN = user_token


def get_user_token():
    return USER_TOKEN


@router.on_event("startup")
async def startup():
    await database.connect()
    if auth_config.mode == "noauth":
        await create_noauth_user()
    elif auth_config.mode == "token":
        set_user_token(str(uuid4()))
        await create_token_user()


@router.on_event("shutdown")
async def shutdown():
    global TOKEN_USER
    await database.disconnect()
    if auth_config.mode == "token":
        await user_db.delete(TOKEN_USER)


async def create_noauth_user():
    noauth_user = await user_db.get_by_email(NOAUTH_EMAIL)
    if noauth_user is None:
        await user_db.create(NOAUTH_USER)


async def create_token_user():
    global TOKEN_USER
    logger.info("To access the server, copy and paste this URL:")
    logger.info(
        f"http://{uvicorn_config.host}:{uvicorn_config.port}/?token={USER_TOKEN}"
    )
    token_email = f"{USER_TOKEN}_user@jupyter.com"
    TOKEN_USER = UserDB(id=USER_TOKEN, email=token_email, hashed_password="")
    await user_db.create(TOKEN_USER)


@router.get("/auth/users")
async def get_users(user: User = Depends(current_user())):
    users = session.query(UserTable).all()
    return [user for user in users if user.logged_in]


r_auth = register_router(auth_router)
r_register = register_router(user_register_router)
r_users = register_router(users_router, prefix="/auth/users")
r_github = register_router(github_oauth_router, prefix="/auth/github")
r = register_router(router)
